<?php
/**
 * @file
 * Defines the default query object which builds FQL to execute using the
 * Facebook's GraphAPI.
 *
 * @todo
 * Figure out an inexpensive method to keep track of item count for pagers;
 * Get rid of count query;
 * Implement query results caching;
 *
 */

/**
 * Object used to create a SELECT query.
 */
class fbapp_viewsfql_plugin_query_graphapi extends views_plugin_query {

  /**
   * An array of sections of the WHERE query. Each section is in itself
   * an array of pieces and a flag as to whether or not it should be AND
   * or OR.
   */
  var $where = array();

  /**
   * A simple array of order by clauses.
   */
  var $orderby = array();

  /**
   * The table header to use for tablesort. This matters because tablesort
   * needs to modify the query and needs the header.
   */
  var $header = array();

  /**
   * The current used pager plugin.
   *
   * @var views_plugin_pager
   */
  var $pager = NULL;

  /**
   * The access token to use if set by an argument.
   */
  var $access_token = NULL;

  /**
   * The actor's data temporary cache.
   */
  var $actors = NULL;

  /**
   * Constructor; Create the basic query object and fill with default values.
   */
  function init($base_table = 'fbapp_user', $base_field = 'uid', $options) {
    // Strip 'fbapp_' prefix from table name.
    $base_table = str_replace('fbapp_', '', $base_table);
    parent::init($base_table, $base_field, $options);
    $this->base_table = $base_table;
    $this->base_field = $base_field;
  }

  /**
   * Set the table header; used for click-sorting because it's needed
   * info to modify the ORDER BY clause.
   */
  function set_header($header) {
    $this->header = $header;
  }

  function option_definition() {
    $options = parent::option_definition();
    $options['use_token'] = array(
      'default' => 'application_token',
    );

    return $options;
  }

  /**
   * Add settings for the ui.
   */
  function options_form(&$form, &$form_state) {
    $form['use_token'] = array(
      '#type' => 'select',
      '#title' => t('Auth token to use'),
      '#description' => t('The authentication token to use when executing the query. Remember that data may be restricted to the entity that owns the token.'),
      '#default_value' => $this->options['use_token'],
      '#options' => array(
        'application_token' => t('Application'),
        'user_token' => t('Logged-in user'),
        'argument_token' => t('User from argument'),
      ),
    );
  }

  /**
   * Add a field to the query table, possibly with an alias. This will
   * automatically call ensure_table to make sure the required table
   * exists, *unless* $table is unset.
   *
   * @param $table
   *   The table this field is attached to. If NULL, it is assumed this will
   *   be a formula; otherwise, ensure_table is used to make sure the
   *   table exists.
   * @param $field
   *   The name of the field to add. This may be a real field or a formula.
   * @param $alias
   *   The alias to create. If not specified, the alias will be $table_$field
   *   unless $table is NULL. When adding formulae, it is recommended that an
   *   alias be used.
   * @param $params
   *   An array of parameters additional to the field that will control items
   *   such as aggregation functions and DISTINCT.
   *
   * @return $name
   *   The name that this field can be referred to as. Usually this is the alias.
   */
  function add_field($table, $field, $alias = '', $params = array()) {
    // We check for this specifically because it gets a special alias.
    if ($table == $this->base_table && $field == $this->base_field && empty($alias)) {
      $alias = $this->base_field;
    }

    if ($table && empty($this->table_queue[$table])) {
      $this->ensure_table($table);
    }

    // Make sure an alias is assigned
    $alias = $alias ? $alias : $field;

    // PostgreSQL truncates aliases to 63 characters: http://drupal.org/node/571548

    // We limit the length of the original alias up to 60 characters
    // to get a unique alias later if its have duplicates
    $alias = drupal_substr($alias, 0, 60);

    // Create a field info array.
    $field_info = array(
      'field' => $field,
      'table' => $table,
      'alias' => $alias,
    ) + $params;

    // Test to see if the field is actually the same or not. Due to
    // differing parameters changing the aggregation function, we need
    // to do some automatic alias collision detection:
    $base = $alias;
    $counter = 0;
    while (!empty($this->fields[$alias]) && $this->fields[$alias] != $field_info) {
      $field_info['alias'] = $alias = $base . '_' . ++$counter;
    }

    if (empty($this->fields[$alias])) {
      $this->fields[$alias] = $field_info;
    }

    return $alias;
  }

  /**
   * Remove all fields that may've been added; primarily used for summary
   * mode where we're changing the query because we didn't get data we needed.
   */
  function clear_fields() {
    $this->fields = array();
  }

  /**
   * Add a simple WHERE clause to the query. The caller is responsible for
   * ensuring that all fields are fully qualified (TABLE.FIELD) and that
   * the table already exists in the query.
   *
   * @param $group
   *   The WHERE group to add these to; groups are used to create AND/OR
   *   sections. Groups cannot be nested. Use 0 as the default group.
   *   If the group does not yet exist it will be created as an AND group.
   * @param $clause
   *   The actual clause to add. When adding a where clause it is important
   *   that all tables are addressed by the alias provided by add_table or
   *   ensure_table and that all fields are addressed by their alias wehn
   *   possible. Please use %d and %s for arguments.
   * @param ...
   *   A number of arguments as used in db_query(). May be many args or one
   *   array full of args.
   */
  function add_where($group, $clause) {
    $args = func_get_args();

    array_shift($args); // ditch $group
    array_shift($args); // ditch $clause

    // Expand an array of args if it came in.
    if (count($args) == 1 && is_array(reset($args))) {
      $args = current($args);
    }

    // Ensure all variants of 0 are actually 0. Thus '', 0 and NULL are all
    // the default group.
    if (empty($group)) {
      $group = 0;
    }

    // Check for a group.
    if (!isset($this->where[$group])) {
      $this->set_where_group('AND', $group);
    }

    // Add the clause and the args.
    if (is_array($args)) {
      $this->where[$group]['clauses'][] = $clause;
      // we use array_values() here to prevent array_merge errors as keys from multiple
      // sources occasionally collide.
      $this->where[$group]['args'] = array_merge($this->where[$group]['args'], array_values($args));
    }
  }

  /**
   * Add an ORDER BY clause to the query.
   *
   * @param $table
   *   The table this field is part of. If a formula, enter NULL.
   * @param $field
   *   The field or formula to sort on. If already a field, enter NULL
   *   and put in the alias.
   * @param $order
   *   Either ASC or DESC.
   * @param $alias
   *   The alias to add the field as. In SQL, all fields in the order by
   *   must also be in the SELECT portion. If an $alias isn't specified
   *   one will be generated for from the $field; however, if the
   *   $field is a formula, this alias will likely fail.
   * @param $params
   *   Any params that should be passed through to the add_field.
   */
  function add_orderby($table, $field, $order, $alias = '', $params = array()) {
    if ($table) {
      $this->ensure_table($table);
    }

    // Only fill out this aliasing if there is a table;
    // otherwise we assume it is a formula.
    if (!$alias && $table) {
      $as = $table . '_' . $field;
    }
    else {
      $as = $alias;
    }

    if ($field) {
      $as = $this->add_field($table, $field, $as, $params);
    }

    $this->orderby[] = "$as " . drupal_strtoupper($order);
  }

  /**
   * Construct the "WHERE" or "HAVING" part of the query.
   *
   * @param $where
   *   'where' or 'having'.
   */
  function condition_sql($where = 'where') {
    $clauses = array();
    foreach ($this->$where as $group => $info) {
      if (empty($info['clauses'])) {
        continue;
      }

      $clause = implode(") " . $info['type'] . " (", $info['clauses']);
      if (count($info['clauses']) > 1) {
        $clause = '(' . $clause . ')';
      }
      $clauses[] = $clause;
    }

    if ($clauses) {
      $keyword = drupal_strtoupper($where);
      if (count($clauses) > 1) {
        return "$keyword (" . implode(") " . $this->group_operator . ' (', array_filter($clauses)) . ")\n";
      }
      else {
        return "$keyword " . array_shift($clauses) . " ";
      }
    }
    return "";
  }

  function compile_fields($fields_array) {
    $fields = array();

    foreach ($fields_array as $field) {
      $string = '';

      $string .= $field['field'];

      if (!empty($field['count'])) {
        // Retained for compatibility
        $field['function'] = 'count';
      }

      if (!empty($field['function'])) {
        $info = $this->get_aggregation_info();
        if (!empty($info[$field['function']]['method']) && function_exists($info[$field['function']]['method'])) {
          $string = $info[$field['function']]['method']($field['function'], $string);
        }
      }

      $fields[] = $string;

      if (!empty($get_count_optimized)) {
        // We only want the first field in this case.
        break;
      }
    }
    return array(
      $fields,
    );
  }

  /**
   * Generate a query and a countquery from all of the information supplied
   * to the object.
   *
   * @param $get_count
   *   Provide a countquery if this is true, otherwise provide a normal query.
   */
  function query($get_count = FALSE) {

    /**
     * An optimized count query includes just the base field instead of all the fields.
     */
    $fields_array = $this->fields;
    if ($get_count) {
      foreach ($fields_array as $field) {
        if (!empty($field['function'])) {
          $get_count_optimized = FALSE;
          break;
        }
      }
    }
    else {
      $get_count_optimized = FALSE;
    }
    if (!isset($get_count_optimized)) {
      $get_count_optimized = TRUE;
    }

    $where = $orderby = '';

    list($fields) = $this->compile_fields($fields_array);

    if (!$get_count_optimized) {
      // We only add the orderby if we're not counting, because of performance reasons.
      if ($this->orderby) {
        $orderby = "ORDER BY " . implode(', ', $this->orderby) . " ";
      }
    }

    $where = $this->condition_sql();

    $query = "SELECT " . implode(', ', $fields) . " FROM $this->base_table $where $orderby";

//    $replace = array('&gt;' => '>', '&lt;' => '<');
//    $query = strtr($query, $replace);

    return $query;
  }

  /**
   * Get the arguments attached to the WHERE and HAVING clauses of this query.
   */
  function get_where_args() {
    $args = array();
    foreach ($this->where as $group => $where) {
      $args = array_merge($args, $where['args']);
    }
    return $args;
  }

  /**
   * Let modules modify the query just prior to finalizing it.
   */
  function alter(&$view) {
    foreach (module_implements('fbapp_views_query_alter') as $module) {
      $function = $module . '_fbapp_views_query_alter';
      $function($view, $this);
    }
  }

  /**
   * Builds the necessary info to execute the query.
   */
  function build(&$view) {
    // Store the view in the object to be able to use it later.
    $this->view = $view;

    $view->init_pager();

    // Let the pager modify the query to add limits.
    $this->pager->query();

    $this->query = $this->query();
    $this->final_query = $this->query;
    $this->count_query = ''; //$this->query(TRUE);
    $this->query_args = $this->get_where_args();
  }

  /**
   * Executes the query and fills the associated view object with according
   * values.
   *
   * Values to set: $view->result, $view->total_rows, $view->execute_time,
   * $view->current_page.
   */
  function execute(&$view) {
    $query = $this->final_query;
    $count_query = $this->count_query;
    $args = $this->query_args;

//    vpr($query);

    if ($query) {
      $replacements = module_invoke_all('views_query_substitutions', $view);
      $query = str_replace(array_keys($replacements), $replacements, $query);
//      $count_query = 'SELECT count() FROM (' . str_replace(array_keys($replacements), $replacements, $count_query) . ')';

      if (is_array($args)) {
        foreach ($args as $id => $arg) {
          $args[$id] = str_replace(array_keys($replacements), $replacements, $arg);
        }
      }

      $start = views_microtime();

      if ($this->pager->use_count_query() || !empty($view->get_total_rows)) {
//        $total_items = $this->get_total_items($query, $args);
        $total_items = 999999;
        $this->pager->total_items = $total_items;
        $this->pager->update_page_info();
        $view->total_rows = $total_items;
      }

      if (!empty($this->limit) || !empty($this->offset)) {
        // We can't have an offset without a limit, so provide a very large limit instead.
        $limit  = intval(!empty($this->limit) ? $this->limit : 999999);
        $offset = intval(!empty($this->offset) ? $this->offset : 0);
        $result = $this->db_query_range($query, $offset, $limit, $args);
      }
      else {
        $result = $this->db_query($query, $args);
      }

      $view->result = array();
      foreach ($result as $item) {
        $view->result[] = $item;
      }

      if ($this->pager->use_pager()) {
        $this->pager->set_current_page($view->current_page);
        $this->pager->total_items = count($result);
      }
    }
    $view->execute_time = views_microtime() - $start;
  }

  function get_total_items($query, $args) {
    $result = $this->db_query($query, $args);
    return count($result);
  }

  function set_access_token($access_token) {
    if ($this->access_token == NULL) {
      $this->access_token = $access_token;
    }
  }

  function get_access_token() {
    switch ($this->options['use_token']) {
      case 'user_token':
        global $user;
        $user_info = fbapp_user_fb_data($user->uid, TRUE);
        $token = $user_info['access_token'];
        break;
      case 'argument_token':
        $token = $this->access_token;
        break;
      default:
        $token = variable_get('fbapp_app_access_token', '');
        break;
    }
    return $token;
  }

  /**
   * Wrapper method for db_query().
   *
   * @param $query
   *   A string containing an SQL query.
   * @param $args
   *   An array containing the query arguments.
   * @return
   *   A database query result resource, or FALSE if the query was not executed
   *   correctly.
   */
  function db_query($query, $args = array()) {
    $token = $this->get_access_token();

    if ($args) {
      _db_query_callback($args, TRUE);
      $query = preg_replace_callback(DB_QUERY_REGEXP, '_db_query_callback', $query);
    }

    return fbapp_fql_query($query, $token);
  }

  /**
   * Wrapper method for db_query_range().
   *
   * @param $query
   *   A string containing an SQL query.
   * @param $from
   *   The first result row to return.
   * @param $count
   *   The maximum number of result rows to return.
   * @param $args
   *   An array containing the query arguments.
   * @return
   *   A database query result resource, or FALSE if the query was not executed
   *   correctly.
   */
  function db_query_range($query, $from, $count, $args = array()) {
    return $this->db_query($query . ' LIMIT ' . $count . ' OFFSET ' . $from, $args);
  }

  /**
   * Wrapper method for db_result().
   *
   * @param $result
   *   A database query result resource, as returned from db_query().
   * @return
   *   The resulting field or FALSE.
   */
  function db_result($result) {
    $data = $this->db_query($result);
    return $data[0][0];
  }

  function add_signature(&$view) {
    $view->query->add_field(NULL, "'" . $view->name . ':' . $view->current_display . "'", 'view_name');
  }

  function get_aggregation_info() {
    return NULL;
  }

  /**
   * Return info to base the uniqueness of the result on.
   *
   * @return $cache_info
   *   A string or array with query unique data or FALSE to deactivate caching
   */
  function get_cache_info() {
    return array(
      'final_query' => $this->final_query,
      'count_query' => $this->count_query,
      'query_args' => $this->query_args,
    );
  }

  /**
   * Return preview info.
   *
   * @return $rows
   *   An array of strings with info to be shown in the preview.
   */
  function get_preview_info() {
    $rows = array();

    $query = $this->query;
    if ($this->query_args) {
      _db_query_callback($this->query_args, TRUE);
      $query = preg_replace_callback(DB_QUERY_REGEXP, '_db_query_callback', $query);
    }
    $rows[] = array('<strong>' . t('FQL query') . '</strong>', '<pre>' . check_plain($query) . '</pre>');

    return $rows;
  }

  function get_actor_data($actor_id) {
    if (!isset($actors[$actor_id])) {
      $params = array(
        'access_token' => $this->get_access_token(),
      );
      $actor_data = fbapp_graph_request($actor_id, $params);
      $actors[$actor_id] = $actor_data;
    }
    return $actors[$actor_id];
  }
}

