<?php

/**
 * @file
 * Functions to access and manipulate data through FB's API.
 *
 * @todo
 * Implement a decent error handling policy;
 *
 */

/**
 * Retrieve application token.
 */
function fbapp_app_authenticate($app_id, $app_secret) {
  $params = array(
    'client_id' => $app_id,
    'client_secret' => $app_secret,
    'grant_type' => 'client_credentials',
  );
  $token = fbapp_get_token($params);
  if ($token) {
    return $token;
  }
  return FALSE;
}

/**
 * Helper function to get access tokens.
 */
function fbapp_get_token($params) {
  $url = url("https://graph.facebook.com/oauth/access_token", array('query' => $params));
  $return = drupal_http_request($url);
  if ($return->code == 200 && isset($return->data)) {
    parse_str($return->data, $tokendata);
    return $tokendata;
  }
  return FALSE;
}

/**
 * Execute Facebook Graph request.
 */
function fbapp_graph_request($target, $parameters, $method = 'GET') {
  // Check cache first.
  $object = fbapp_cache_pre_request($target, $method);
  if ($object != 0) {
    return $object;
  }
  // We need to use CURL because drupal_http_request() isn't great to post files.
  $ch = curl_init();
  $target_url = "https://graph.facebook.com/{$target}";
  switch ($method) {
    case 'POST':
      curl_setopt($ch, CURLOPT_POST, 1);
      curl_setopt($ch, CURLOPT_POSTFIELDS, $parameters);
      break;
    case 'DELETE':
      curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "DELETE");
      curl_setopt($ch, CURLOPT_POSTFIELDS, $parameters);
      break;
    default: // 'GET'
      $target_url = url($target_url, array('query' => $parameters));
      break;
  }
  curl_setopt($ch, CURLOPT_URL, $target_url);
  curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
  $return = curl_exec($ch);
  //dpm($return);
  $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
  if (curl_errno($ch)) {
    error_log('Curl error: ' . curl_error($ch));
  }
  curl_close($ch);
  if ($http_code == 200) {
    $object = json_decode($return, TRUE);
    // Cache/update cached object.
    fbapp_cache_post_request($target, $method, $object);
    return $object;
  }
  return FALSE;
}

/**
 * Execute FQL query.
 */
function fbapp_fql_query($query, $access_token) {
  $parameters = array(
    'access_token' => $access_token,
    'format' => 'json',
    'query' => $query,
  );
  $url = url('https://api.facebook.com/method/fql.query', array('query' => $parameters));
  $return = drupal_http_request($url);
  $data = array();
  if ($return->code == 200) {
    $data = json_decode($return->data);
  }
  if (!is_array($data) || isset($data['error_code'])) {
    dpm($query);
    dpm($return);
  }
  return $data;
}

/**
 * Reads the application properties from Facebook and saves them in variables.
 * Reference: http://developers.facebook.com/docs/reference/rest/admin.getAppProperties/
 */
function fbapp_app_properties_get() {
  $params = array(
    'access_token' => variable_get('fbapp_app_access_token', ''),
    'properties' => implode(',', fbapp_app_properties_list()),
    'format' => 'json',
  );
  // Read settings from FB.
  $url = url("https://api.facebook.com/method/admin.getAppProperties", array('query' => $params));
  $return = drupal_http_request($url);
  if ($return->code == 200 && isset($return->data)) {
    // Looks stupid, but apparently Facebook double-json-encodes this data. (???)
    $app_props = json_decode($return->data, TRUE);
    $app_props = json_decode($app_props, TRUE);
    // Save settings to local vars.
    foreach ($app_props as $key => $value) {
      variable_set('fbapp_' . $key, $value);
    }
    return TRUE;
  }
  return FALSE;
}

/**
 * Writes the application properties to Facebook.
 * Reference: http://developers.facebook.com/docs/reference/rest/admin.setAppProperties/
 */
function fbapp_app_properties_set($properties) {
  $params = array(
    'access_token' => variable_get('fbapp_app_access_token', ''),
    'properties' => json_encode($properties),
    'format' => 'json',
  );
  $url = url("https://api.facebook.com/method/admin.setAppProperties", array('query' => $params));
  $return = drupal_http_request($url);
  if ($return->data == 'true') {
    return TRUE;
  }
  return FALSE;
}

/**
 * Decodes a signed request sent by Facebook.
 * Reference: https://developers.facebook.com/docs/authentication/signed_request/
 */
function fbapp_parse_signed_request($signed_request, $secret) {
  list($encoded_sig, $payload) = explode('.', $signed_request, 2);

  // Decode the data.
  $sig = fbapp_base64_url_decode($encoded_sig);
  $data = json_decode(fbapp_base64_url_decode($payload), TRUE);

  if (drupal_strtoupper($data['algorithm']) !== 'HMAC-SHA256') {
    watchdog('fbapp', 'Unknown algorithm detected parsing signed request.', WATCHDOG_ERROR);
    return NULL;
  }

  // Check signature.
  $expected_sig = hash_hmac('sha256', $payload, $secret, $raw = TRUE);
  if ($sig !== $expected_sig) {
    watchdog('fbapp', 'Bad JSON signature detected parsing signed request.', WATCHDOG_ERROR);
    return NULL;
  }

  return $data;
}

/**
 * Base64-decodes an URL.
 * Reference: https://developers.facebook.com/docs/authentication/signed_request/
 */
function fbapp_base64_url_decode($input) {
  return base64_decode(strtr($input, '-_', '+/'));
}

